try:
    from .events import Events
    from .async_events import AsyncEvents
    from .objects import Parameters, Decorator, Event
    from .simple_events import EasyEvents
except ImportError:
    from events import Events
    from async_events import AsyncEvents
    from objects import Parameters, Decorator, Event
    from simple_events import EasyEvents



# coding: utf-8
from ast import literal_eval
from asyncio import iscoroutinefunction


class Parameters:
    def __init__(self, data, prefix: str = "", str_only: bool = True):
        self._prefix = prefix
        self._called = True
        self._event = data
        self._parameters = ""

        if not str_only:
            self.revert()
        else:
            self.convert()

    def convert(self):
        check = False
        blank_prefix = False
        com = self._event

        if isinstance(self._prefix, str):
            check = str(self._event).startswith(self._prefix)
            blank_prefix = False if self._prefix else True

        elif isinstance(self._prefix, list):
            for pre in self._prefix:
                if str(self._event).startswith(pre):
                    check = True

                if check and pre == "":
                    blank_prefix = True

        if check and not blank_prefix:
            try:
                self._event = com.lower().split()[0][1:]
                self._parameters = " ".join(com.split()[1:])
            except Exception:
                self._event = com.lower()
                self._parameters = ""

        elif check and blank_prefix:
            self._event = com.lower().split()[0]
            self._parameters = " ".join(com.split()[1:])

        self._called = check

    def revert(self):
        done = False
        try:
            data = {"event": "", "parameters": ""}

            if isinstance(self._event, str) or isinstance(self._event, bytes):
                data = literal_eval(self._event)

            elif isinstance(self._event, list):
                if self._event:
                    data["event"] = self._event[0]

                if len(self._event) > 1:
                    data["parameters"] = self._event[1:]

            else:
                data = self._event

            self._event = data.get("event")
            self._parameters = data.get("parameters")

            done = True

            for key, value in data.items():
                if key not in ["event", "parameters"]:
                    setattr(self, key, value)

        except Exception:
            if not done:
                self.convert()

    def transform(self):
        return self.__dict__

    def clear(self):
        keys = [k for k in self.__dict__.keys()]
        for key in keys:
            delattr(self, key)

    def clean(self):
        del(self._event)
        del(self._parameters)
        del(self._prefix)
        del(self._called)

    def build_str(self):
        res = ""
        for key, value in self.__dict__.items():
            res += f"{key} : {value}\n"

        return res

    def get(self, key: str):
        getattr(self, key, None)

    def __str__(self):
        return self.build_str()

    def setattr(self, key, value):
        setattr(self, key, value)

    def delattr(self, key):
        delattr(self, key)


class Event:
    def __init__(self,
                 names: list,
                 event: callable,
                 condition: callable = None,
                 event_type: str = None
                 ):

        self.names = names
        self.event = event
        self.condition = condition
        self.type = event_type

    def check_type(self, event_type: str = None):
        if not event_type:
            return True

        return self.type == event_type


class Decorator:
    def __init__(self, is_async: bool = False, use_funct_name: bool = True, default_event: bool=True):
        self.events = []
        self.is_async = is_async
        self.use_funct_name = use_funct_name

        if default_event:
            self.event = self.add_event

    def event_exist(self, name: str):
        return name in self.get_all_events_names()

    def type_exist(self, name: str):
        return name in self.get_types()

    def get_all_events_names(self):
        liste = []
        for ev in self.events:
            for name in ev.names:
                liste.append(name)

        return liste

    def get_all_events_types(self):
        liste = []
        for ev in self.events:
            liste.append(ev.type)

        return liste

    def get_events_names(self, event_type: str = None):
        liste = []
        for ev in self.events:
            for name in ev.names:
                if ev.type == event_type:
                    liste.append(name)

        return liste

    def get_events_type(self, event_type: str = None):
        liste = []
        for event in self.events:
            if event_type == event.type:
                liste.append(event)

        return liste

    def get_types(self):
        liste = []
        for ev in self.events:
            liste.append(ev.type)

        liste = list(dict.fromkeys(liste))
        return liste

    def get_event(self, name: str):
        for event in self.events:
            if name in event.names:
                return event

    def grab_event(self, name: str, event_type: str = None):
        for event in self.events:
            if name in event.names and event_type == event.type:
                return event

    def remove_event(self, name: str, event_type: str = None):
        event = self.grab_event(name, event_type)
        if event:
            self.events.remove(event)

    def add_event(self, aliases: list = [], condition: callable = None, type: str = None, callback: callable = None, event_type: str = None):
        if isinstance(aliases, str):
            aliases = [aliases]

        if not callable(condition):
            condition = None

        if type:
            event_type = type

        def add_command(command_funct):
            if self.is_async and not iscoroutinefunction(command_funct):
                raise '[Error] Command must be async: "async def ..."'

            if self.use_funct_name:
                aliases.append(command_funct.__name__)

            al = list(dict.fromkeys(aliases))

            self.events.append(Event(al, command_funct, condition, event_type))
            return command_funct

        if callable(callback):
            return add_command(callback)

        return add_command



# coding: utf-8
from inspect import getfullargspec
from threading import Thread
import time
try:
    from .objects import Decorator, Parameters, Event
except ImportError:
    from objects import Decorator, Parameters, Event

# import json
# from types import SimpleNamespace
# x = json.loads(data, object_hook=lambda _d: SimpleNamespace(**_d))


class Events(Decorator):
    def __init__(self,
                 prefix: str = "",
                 str_only: bool = False,
                 use_funct_name: bool = True,
                 first_parameter_object: bool = True,
                 default_event: bool = True
                 ):

        Decorator.__init__(self, is_async=False, use_funct_name=use_funct_name, default_event=default_event)
        self.prefix = prefix
        self._str_only = str_only
        self.process_data = self.trigger
        self.waiting_list = []
        self.first_parameter_object = first_parameter_object

    def build_arguments(self, function, arguments):
        values = getfullargspec(function)

        arg = values.args

        default = values.defaults
        ext_default = values.kwonlydefaults
        ext = None

        if not arg and not values.kwonlyargs:
            return None

        if self.first_parameter_object:
            arg.pop(0)

        if values.kwonlyargs:
            ext = values.kwonlyargs[0]
            arg.extend(values.kwonlyargs)

        para = {}

        if default:
            default = list(default)
            for i in range(-1, -len(default)-1, -1):
                para[arg[i]] = default[i]

        s = len(arg)
        dico = {}

        if ext:
            if not (isinstance(arguments, list) or isinstance(arguments, dict)):
                arguments = arguments.split()

            sep = len(arguments) - s + 1

            if not sep:
                sep = 1

            for i in range(s):
                key = arg[i]

                if key != ext:
                    if isinstance(arguments, list):
                        try:
                            dico[key] = arguments.pop(0)
                        except IndexError:
                            if key in para.keys():
                                dico[key] = para[key]

                    elif isinstance(arguments, dict):
                        try:
                            dico[key] = arguments[key]
                        except KeyError:
                            if key in para.keys():
                                dico[key] = para[key]

                else:
                    li = []
                    if isinstance(arguments, list):
                        for _ in range(sep):
                            try:
                                li.append(arguments.pop(0))
                            except IndexError:
                                pass

                    elif isinstance(arguments, dict):
                        try:
                            li = arguments[key]
                        except KeyError:
                            pass

                    if not li and ext_default and ext_default.get(key):
                        li = [ext_default[key]]

                    dico[key] = li

        elif s:
            if isinstance(arguments, list):
                dico = {key: value for key, value in zip(arg, arguments[0:s])}

            elif isinstance(arguments, dict):
                for key in arg:
                    try:
                        dico[key] = arguments[key]
                    except KeyError:
                        if key in para.keys():
                            dico[key] = para[key]
            else:
                dico = {key: value for key, value in zip(arg, arguments.split()[0:s])}

        return dico

    def execute(self, event: Event, data: Parameters):
        com = event.event
        con = event.condition

        dico = self.build_arguments(com, data._parameters)

        for elem in ["_event", "_parameters", "_prefix", "_called"]:
            delattr(data, elem)

        if (con and con(data)) or not con:
            if not isinstance(dico, dict):
                return com()

            if self.first_parameter_object:
                return com(data, **dico)

            return com(**dico)

    def trigger(self, data, event_type: str = None, str_only: bool = None):
        none = type(None)

        if isinstance(str_only, none):
            str_only = self._str_only

        args = data

        if isinstance(data, Parameters):
            pass
        elif not str(type(data)) == "<class 'easy_events.objects.Parameters'>":
            args = Parameters(data, self.prefix, str_only)

        event = self.grab_event(args._event, event_type)

        if isinstance(args._event, str) and event and args._called:
            try:
                val = self.execute(event, args)
            except Exception as e:
                raise e
                return f"{type(e)}: {e}"

            if isinstance(val, Parameters):
                return val.transform()

            return val

        return data

    def add_task(self, data, event_type: str = None, str_only: bool = None):
        none = type(None)

        if isinstance(str_only, none):
            str_only = self._str_only

        args = data

        if isinstance(data, Parameters):
            pass
        elif not str(type(data)) == "<class 'easy_events.objects.Parameters'>":
            args = Parameters(data, self.prefix, str_only)

        event = self.grab_event(args._event, event_type)

        if isinstance(args._event, str) and event and args._called:
            self.waiting_list.append((event, args))

    def run_task(self, thread: bool = False):
        for data in self.waiting_list:
            if thread:
                Thread(target=self.execute, args=[*data]).start()
            else:
                self.execute(*data)

        self.waiting_list.clear()


if __name__ == "__main__":
    client = Events(first_parameter_object=False, str_only=False)

    @client.event()
    def test1(arg1, arg2="", *, arg3=""):
        print("test1", arg1, arg2, arg3)
        print("data")

    # @client.event()
    def test2():
        print("test2") # , arg1, arg2, arg3)
        print("data") # , data)

    # client.event(aliases="event_name", type="event", callback=test1)
    # client.event(callback=test2)

    client.trigger("test1 1 2 3")
    print("-"*50)

    client.trigger({"event": "test1", "parameters": {"arg1": "a1", "arg2": "a2", "arg3": ["a1", "a2"]}})
    client.trigger({"event": "test1", "parameters": [1, 2, 3, 4, 5]})

    client.trigger({"event": "test2", "parameters": ["arg1", "arg2", "arg3", "arg4"]})
    # client.trigger(Parameters("test1"))

    data = Parameters("event_name")
    data.client = "hello"
    client.trigger(data)



from inspect import getfullargspec
from threading import Thread
from asyncio import (
    gather,
    wait_for,
    new_event_loop,
    iscoroutinefunction,
    run_coroutine_threadsafe,
    AbstractEventLoop,
    TimeoutError,
    Future
)
import asyncio

try:
    from .objects import Decorator, Parameters, Event
except ImportError:
    from objects import Decorator, Parameters, Event

# import json
# from types import SimpleNamespace
# x = json.loads(data, object_hook=lambda _d: SimpleNamespace(**_d))


class AsyncEvents(Decorator):
    def __init__(self,
                 prefix: str = "",
                 str_only: bool = False,
                 use_funct_name: bool = True,
                 first_parameter_object: bool = True,
                 default_event: bool = True
                 ):
        Decorator.__init__(self, is_async=True, use_funct_name=use_funct_name, default_event=default_event)
        self.prefix = prefix
        self._str_only = str_only
        self._run = True
        # self._loop: AbstractEventLoop = new_event_loop()
        self.waiting_list = []
        self.process_data = self.trigger
        self.first_parameter_object = first_parameter_object

    async def build_arguments(self, function, arguments):
        values = getfullargspec(function)

        arg = values.args

        default = values.defaults
        ext_default = values.kwonlydefaults
        ext = None

        if not arg and not values.kwonlyargs:
            return None

        if self.first_parameter_object:
            arg.pop(0)

        if values.kwonlyargs:
            ext = values.kwonlyargs[0]
            arg.extend(values.kwonlyargs)

        para = {}

        if default:
            default = list(default)
            for i in range(-1, -len(default)-1, -1):
                para[arg[i]] = default[i]

        s = len(arg)
        dico = {}

        if ext:
            if not (isinstance(arguments, list) or isinstance(arguments, dict)):
                arguments = arguments.split()

            sep = len(arguments) - s + 1

            if not sep:
                sep = 1

            for i in range(s):
                key = arg[i]

                if key != ext:
                    if isinstance(arguments, list):
                        try:
                            dico[key] = arguments.pop(0)
                        except IndexError:
                            if key in para.keys():
                                dico[key] = para[key]

                    elif isinstance(arguments, dict):
                        try:
                            dico[key] = arguments[key]
                        except KeyError:
                            if key in para.keys():
                                dico[key] = para[key]

                else:
                    li = []
                    if isinstance(arguments, list):
                        for _ in range(sep):
                            try:
                                li.append(arguments.pop(0))
                            except IndexError:
                                pass

                    elif isinstance(arguments, dict):
                        try:
                            li = arguments[key]
                        except KeyError:
                            pass

                    if not li and ext_default and ext_default.get(key):
                        li = [ext_default[key]]

                    dico[key] = li

        elif s:
            if isinstance(arguments, list):
                dico = {key: value for key, value in zip(arg, arguments[0:s])}

            elif isinstance(arguments, dict):
                for key in arg:
                    try:
                        dico[key] = arguments[key]
                    except KeyError:
                        if key in para.keys():
                            dico[key] = para[key]
            else:
                dico = {key: value for key, value in zip(arg, arguments.split()[0:s])}

        return dico

    async def execute(self, event: Event, data: Parameters):
        com = event.event
        con = event.condition

        dico = await self.build_arguments(com, data._parameters)

        data._parameters = dico

        if (con and con(data)) or not con:
            if not isinstance(dico, dict):
                return await com()

            if self.first_parameter_object:
                return await com(data, **dico)

            return await com(**dico)

    async def trigger(self, data, event_type: str = None, str_only: bool = None):
        none = type(None)

        if isinstance(str_only, none):
            str_only = self._str_only

        args = data

        if isinstance(data, Parameters):
            pass
        elif not str(type(data)) == "<class 'easy_events.objects.Parameters'>":
            args = Parameters(data, self.prefix, str_only)

        event = self.grab_event(args._event, event_type)

        if isinstance(args._event, str) and event and args._called:
            return await self.execute(event, args)

    def add_task(self, data, event_type: str = None, str_only: bool = None):
        none = type(None)

        if isinstance(str_only, none):
            str_only = self._str_only

        args = data

        if isinstance(data, Parameters):
            pass
        elif not str(type(data)) == "<class 'easy_events.objects.Parameters'>":
            args = Parameters(data, self.prefix, str_only)

        event = self.grab_event(args._event, event_type)

        if isinstance(args._event, str) and event and args._called:
            self.waiting_list.append((event, args))

    async def run_task(self):
        tasks = []

        for data in self.waiting_list:
            tasks.append(asyncio.create_task(self.execute(*data)))

        for task in tasks:
            await task

        self.waiting_list.clear()

    def run_task_sync(self, thread: bool = False):
        if thread:
            Thread(target=self._exec).start()
        else:
            self._exec()

    def _exec(self):
        asyncio.run(self.run_task())

    async def _thread(self):
        while self._run:
            await self.run_task()
            await asyncio.sleep(.1)

    def run(self):
        asyncio.run(self._thread())


if __name__ == "__main__":

    client = AsyncEvents(first_parameter_object=False)

    @client.event()
    async def hello(*, world):
        # await asyncio.sleep(1)
        print(f"Hello {world}")
        return f"Hello {world}"

    def build_data(data):
        data.image = "png"
        data.file = "txt"


    data = Parameters("hello", client.prefix)
    build_data(data)
    data = client.add_task(data)
    # print(0, data)

    data = client.add_task({"event": "hello", "parameters": {"world": "world", "lol": "data"}})
    # print(1, data)

    data = client.add_task({"event": "hello", "parameters": ["world", "data"]})
    # print(2, data)

    data = client.add_task(["hello", "world", "data"])
    # print(3, data)

    data = client.add_task("hello world data4")
    # print(4, data)

    client.run_task_sync()
